@@ -78,17 +78,7 @@ function(x, nrow = 1, ncol = 1, eps = .Machine$double.eps){
 "as.matrix.coo" <- function(x, nrow = 1, ncol = 1, eps = .Machine$double.eps)
 {
 	if (is.matrix.coo(x) && missing(nrow) && missing(ncol)) x
-# the following wouldn't be needed if the lazy evaluation works
-#	else if(!missing(nrow) && !missing(ncol)) 
-#		as.matrix.coo(as.matrix.csr(x, nrow, ncol, eps))
-#	else if(!missing(nrow))
-#		as.matrix.coo(as.matrix.csr(x, nrow = nrow, eps = eps))
-#	else if(!missing(ncol))
-#		as.matrix.coo(as.matrix.csr(x, ncol = ncol, eps = eps))
-#	else
-#		as.matrix.coo(as.matrix.csr(x, eps = eps))
-	else as.matrix.coo(as.matrix.csr(x, nrow, ncol, eps = .Machine$double.eps))
-#	else as.matrix.coo(as.matrix.csr(x, nrow = 1, ncol = 1, eps = .Machine$double.eps))
+	else as.matrix.coo(as.matrix.csr(x))
 }
 #--------------------------------------------------------------------
 "as.matrix.csr.matrix.csc" <- function(x, nrow = 1, ncol = 1, eps = .Machine$double.eps)
@@ -1201,6 +1191,23 @@ else{
 	}
 return(z)
 }
+#--------------------------------------------------------------------
+".kronecker.matrix.csr" <-
+function(X,Y,FUN = "missing", make.dimnames = "missing"){
+	X = as.matrix.coo(X)
+	Y = as.matrix.coo(Y)
+
+	la = length(X@ra)
+	lb = length(Y@ra)
+
+	ra = rep(Y@ra,la)*rep(X@ra,each=lb)
+	ja = as.integer(rep(Y@ja,la)+rep((X@ja-1)*dim(Y)[2],each=lb))
+	ia = as.integer(rep(Y@ia,la)+rep((X@ia-1)*dim(Y)[1],each=lb))
+	dim = as.integer(dim(X)*dim(Y))
+
+	as.matrix.csr(new("matrix.coo",ra=ra,ia=ia,ja =ja,dim=dim))
+}
+
 #--------------------------------------------------------------------
 "chol" <- function(x, ...) UseMethod("chol")
 #--------------------------------------------------------------------
@@ -1366,11 +1373,13 @@ function (x = 1, nrow, ncol=n)
 "diag.assign.matrix.csr" <- function(x,value)
 {
      dx <- dim(x)
-     if (length(dx) != 2 || prod(dx) != length(x))
+#     if (length(dx) != 2 || prod(dx) != length(x))
+     if (length(dx) != 2)
          stop("only matrix diagonals can be replaced")
      i <- seq(length = min(dx))
      if (length(value) != 1 && length(value) != length(i))
          stop("replacement diagonal has wrong length")
+     if (length(value) == 1) value <- rep(value,min(dx))
      x[cbind(i, i)] <- value
      x
 }
@@ -1463,10 +1472,10 @@ function (x, y, ...)
         stop("x and y don't match n")
     chol <- chol(t(x)%*%x)
     xy <- t(x) %*% y
-    coefficients <- backsolve(chol,xy)
-    fitted <-  as.matrix(x %*% coefficients)
-    residuals <- y - fitted
-    return(coefficients, chol, residuals, fitted)
+    coef <- backsolve(chol,xy)
+    fitted <-  as.matrix(x %*% coef)
+    resid <- y - fitted
+    list(coefficients=coef, chol=chol, residuals=resid,  fitted=fitted)
 }
 #--------------------------------------------------------------------
 "coef.slm" <-
@@ -1644,3 +1653,307 @@ function (x, digits = max(3, getOption("digits") - 3),
     cat("\n")
     invisible(x)
 }
+"[.matrix.coo" <- function (x, rw = 1:x@dimension[1], cl = 1:x@dimension[2])
+{
+	nrow <-  x@dimension[1]
+	ncol <-  x@dimension[2]
+	if (is.matrix.coo(rw)) {
+		stop("Indexing by 'matrix.coo' matrices not implemented")
+		}
+	if (is.matrix.csr(rw)) {
+		if(nrow != rw@dimension[1] || ncol != rw@dimension[2])
+			stop("Dimension of the indexing matrix is not the same as the matrix being indexed")
+		x <- as.matrix.coo(t(as.matrix.csr(x)))
+		rw <- as.matrix.coo(t(rw))
+		s <- match(paste(rw@ia,rw@ja),paste(x@ia,x@ja))
+		ra <- x@ra[s]
+		A <- ra
+	        }
+	else if(is.matrix(rw)){
+		if(!all(abs(rw[,1])%in%1:nrow)||!all(abs(rw[,2])%in%1:ncol))
+			stop("Subscripts out of bound")
+		s <- match(paste(rw[,1],rw[,2]),paste(x@ia,x@ja))
+		ra <- x@ra[s]
+		ra[is.na(ra)] <- 0
+		A <- ra
+		}
+	else{
+		if (is.logical(rw))
+			if (length(rw) > nrow)
+				stop("logical subscript too long")
+			else
+				rw <- (1:nrow)[rw]
+		if (is.logical(cl))
+			if (length(cl) > ncol)
+				stop("logical subscript too long")
+			else
+				cl <- (1:ncol)[cl]
+       		if (!all(abs(rw) %in% 1:nrow) || !all(abs(cl) %in% 1:ncol))
+			stop("Subscripts out of bound")
+	        if (any(rw < 0)) {
+			if (!all(rw <= 0))
+		                stop("Only 0's may mix with negative subscripts")
+			else rw <- setdiff(1:nrow, abs(rw))
+		}
+		if (any(cl < 0)) {
+			if (!all(cl <= 0))
+				stop("Only 0's may mix with negative subscripts")
+			else cl <- setdiff(1:ncol, abs(cl))
+		}
+		s <- ((x@ja %in% cl) & (x@ia %in% rw))
+		ja <- match(x@ja[s],cl)
+		ia <- match(x@ia[s],rw)
+		dim <- c(length(rw),length(cl))
+		ra <- x@ra[s]
+		if (length(ra) == 0 && length(ja) == 0){ #trap all zeros returned matrix
+			ra = 0
+			ja = ia = as.integer(1)
+			}
+		A <- new("matrix.coo",ra=ra,ja=ja,ia=ia,dimension=dim)
+	}
+	return(A)
+}
+"[<-.matrix.coo" <-
+function (x, rw = 1:x@dimension[1], cl = 1:x@dimension[2], value) 
+{
+    nrow <-  x@dimension[1]
+    ncol <-  x@dimension[2]
+#    if (!is.matrix.coo(value)) 
+#	 stop("replacement matrix must be of matrix.coo class")
+    if (is.matrix.coo(rw)) {
+         stop("Indexing by 'matrix.coo' matrices not implemented")
+	}
+    else if (is.matrix.csr(rw)) {
+	if(nrow != rw@dimension[1] || ncol != rw@dimension[2])
+		stop("Dimension of the indexing matrix is not the same as the matrix being indexed")
+	x <- as.matrix.coo(t(as.matrix.csr(x)))
+	rw <- as.matrix.coo(t(rw))
+	s <- match(paste(rw@ia,rw@ja),paste(x@ia,x@ja))
+	len.s <- length(s)
+	value <- as.matrix.csr(value)
+	len.value <- length(value@ra)
+	value.ra <- value@ra
+	if (len.s != len.value){
+		if (len.value == 1) 
+			value.ra <- rep(value@ra,len.s)
+		else{
+			if (len.value >= len.s){
+				value.ra <- value@ra[1:len.s]
+				warning("number of items to replace is not a multiple of replacement length")
+				}
+			else {
+				value.ra <- rep(value@ra,len.s%/%len.value+1)[1:len.s]
+				warning("number of items to replace is not a multiple of replacement length")
+				}
+			}
+		}
+	ra <- c(x@ra[-s],value.ra)
+        ia <- as.integer(c(x@ja[-s],rw@ja))
+        ja <- as.integer(c(x@ia[-s],rw@ia))
+        dim <- rev(x@dimension)
+        x <- new("matrix.coo",ra=ra,ja=ja,ia=ia,dimension=dim)
+        }
+    else if(is.matrix(rw)){
+	if(!all(abs(rw[,1])%in%1:nrow)||!all(abs(rw[,2])%in%1:ncol))
+		stop("Subscripts out of bound")
+	s <- match(paste(rw[,1],rw[,2]),paste(x@ia,x@ja))
+#	ra <- c(x@ra[-s],value@ra[!is.na(s)])
+#        ja <- as.integer(c(x@ja[-s],rw[,2][!is.na(s)]))
+#        ia <- as.integer(c(x@ia[-s],rw[,1][!is.na(s)]))
+	value <- as.matrix.csr(value)
+	ra <- c(x@ra[-s],value@ra)
+        ja <- as.integer(c(x@ja[-s],rw[,2]))
+        ia <- as.integer(c(x@ia[-s],rw[,1]))
+        dim <- x@dimension
+        x <- new("matrix.coo",ra=ra,ja=ja,ia=ia,dimension=dim)
+	}
+    else {
+	if (is.logical(rw))
+		if (length(rw) > nrow)
+			stop("logical subscript too long")
+		else
+			rw <- (1:nrow)[rw]
+	if (is.logical(cl))
+		if (length(cl) > ncol)
+			stop("logical subscript too long")
+		else
+			cl <- (1:ncol)[cl]
+        if (!all(abs(rw) %in% 1:nrow) || !all(abs(cl) %in% 1:ncol))
+            stop("Subscripts out of bound")
+        if (any(rw < 0)) {
+            if (!all(rw <= 0))
+                stop("Only 0's may mix with negative subscripts")
+            else rw <- setdiff(1:nrow, abs(rw))
+        }
+        if (any(cl < 0)) {
+            if (!all(cl <= 0))
+                stop("Only 0's may mix with negative subscripts")
+            else cl <- setdiff(1:ncol, abs(cl))
+        }
+        s <- ((x@ja %in% cl) & (x@ia %in% rw))
+	value <- as.matrix.coo(as.matrix.csr(value,nrow=length(rw),ncol=length(cl)))
+        ra <- c(x@ra[!s],value@ra)
+        ja <- as.integer(c(x@ja[!s],cl[value@ja]))
+        ia <- as.integer(c(x@ia[!s],rw[value@ia]))
+        dim <- x@dimension
+        x <- new("matrix.coo",ra=ra,ja=ja,ia=ia,dimension=dim)
+        }
+return(x)
+}
+# All the S4 Methods stuff is collected below this point
+require(methods)
+setClass("matrix.csr",representation(ra="numeric",
+	ja="integer",ia="integer", dimension="integer"),
+	validity = function(object) {
+ 		if(!length(object@dimension) == 2 )
+                	return("invalid dimension attribute")
+        	else{
+               		nrow <- object@dimension[1]
+               		ncol <- object@dimension[2]
+               	 	}
+        	if(!(length(object@ra) ==length(object@ja)))
+                	return("ra and ja don't have equal lengths")
+        	if(any(object@ja < 1) || any(object@ja > ncol)) 
+                	return("ja exceeds dim bounds")
+        	if(any(object@ia < 1))
+                	return("some elements of ia are <= 0")
+		if(any(diff(object@ia)<0))
+			return("ia vector not monotone increasing") 
+		if(object@ia[length(object@ia)] != length(object@ra)+1)
+			return("last element of ia doesn't conform")
+		if(length(object@ia) != nrow+1)
+			return("ia has wrong number of elments")
+        	if(length(object@ra) < 1 || length(object@ra) > nrow*ncol)
+                	return("ra has too few, or too many elements")
+		TRUE})
+setMethod("initialize", "matrix.csr", 
+	function(.Object, ra = numeric(0),  ja = integer(0),
+		ia = integer(0),dimension = integer(0)) {
+        .Object@ra <- ra
+        .Object@ja <- ja
+        .Object@ia <- ia
+        .Object@dimension <- dimension
+	validObject(.Object)
+        .Object
+       })
+setClass("matrix.csc",representation(ra="numeric",ja="integer",ia="integer", dimension="integer"))
+setClass("matrix.ssr",representation(ra="numeric",ja="integer",ia="integer", dimension="integer"))
+setClass("matrix.ssc",representation(ra="numeric",ja="integer",ia="integer", dimension="integer"))
+setClass("matrix.coo",representation(ra="numeric",
+	ja="integer",ia="integer", dimension="integer"),
+	validity = function(object) {
+ 		if(!length(object@dimension) == 2 )
+                	return("invalid dimension attribute")
+        	else{
+               		nrow <- object@dimension[1]
+               		ncol <- object@dimension[2]
+               	 	}
+        	if(!(length(object@ra) ==length(object@ja) && length(object@ra) ==length(object@ia)))
+                	return("ra,ja,ia don't have equal lengths")
+        	if(any(object@ja < 1) || any(object@ja > ncol)) 
+                	return("ja exceeds dim bounds")
+        	if(any(object@ia < 1) || any(object@ia > nrow))
+                	return("ia exceeds dim bounds")
+        	if(length(object@ra) < 1 || length(object@ra) > nrow*ncol)
+                	return("ra has too few, or too many elements")
+		TRUE})
+setMethod("initialize", "matrix.coo", 
+	function(.Object, ra = numeric(0),  ja = integer(0),
+		ia = integer(0),dimension = integer(0)) {
+        .Object@ra <- ra
+        .Object@ja <- ja
+        .Object@ia <- ia
+        .Object@dimension <- dimension
+	validObject(.Object)
+        .Object
+       })
+#-------------------------------------------------------------------------
+setClass("matrix.csr.chol",representation(nrow="numeric",nnzlindx="numeric",
+	nsuper="numeric",lindx="numeric",xlindx="numeric",nnzl="numeric",
+	lnz="numeric",xlnz="numeric",invp="numeric",perm="numeric",
+	xsuper="numeric",ierr="numeric",time="numeric"))
+if(R.version$major >= 1 && R.version$minor >= 8.0){
+	setClassUnion("numeric or NULL",c("numeric","NULL"))
+	setClassUnion("character or NULL",c("character","NULL"))
+	}
+if(R.version$major >= 1 && R.version$minor < 8.0){
+	setClass("numeric or NULL")
+	setIs("numeric","numeric or NULL")
+	setIs("NULL","numeric or NULL")
+	setClass("character or NULL")
+	setIs("character","character or NULL")
+	setIs("NULL","character or NULL")
+	}
+setClass("matrix.csc.hb",representation(ra="numeric",ja="integer",ia="integer", 
+	rhs.ra="numeric",guess="numeric or NULL",xexact="numeric or NULL",dimension ="integer",
+	rhs.dim="numeric",rhs.mode="character or NULL"))
+setClass("matrix.ssc.hb","matrix.csc.hb")
+setClass("slm",representation(coefficients="numeric",chol="matrix.csr.chol",
+	residuals="numeric",fitted="numeric"))
+setClass("mslm","slm")
+setClass("summary.slm","slm")
+#--------------------------------------------------------------------
+setGeneric("as.matrix.csr")
+setMethod("as.matrix.csr","matrix.csc",as.matrix.csr.matrix.csc)
+setMethod("as.matrix.csr","matrix.ssr",as.matrix.csr.matrix.ssr)
+setMethod("as.matrix.csr","matrix.ssc",as.matrix.csr.matrix.ssc)
+setMethod("as.matrix.csr","matrix.coo",as.matrix.csr.matrix.coo)
+setGeneric("as.matrix.csc")
+setMethod("as.matrix.csc","matrix.csr",as.matrix.csc.matrix.csr)
+setMethod("as.matrix.csc","matrix.ssr",as.matrix.csc.matrix.ssr)
+setMethod("as.matrix.csc","matrix.ssc",as.matrix.csc.matrix.ssc)
+setGeneric("as.matrix.ssr")
+setMethod("as.matrix.ssr","matrix.csr",as.matrix.ssr.matrix.csr)
+setMethod("as.matrix.ssr","matrix.csc",as.matrix.ssr.matrix.csc)
+setMethod("as.matrix.ssr","matrix.ssc",as.matrix.ssr.matrix.ssc)
+setGeneric("as.matrix.ssc")
+setMethod("as.matrix.ssc","matrix.csr",as.matrix.ssc.matrix.csr)
+setMethod("as.matrix.ssc","matrix.csc",as.matrix.ssc.matrix.csc)
+setMethod("as.matrix.ssc","matrix.ssr",as.matrix.ssc.matrix.ssr)
+setGeneric("as.matrix.coo")
+setMethod("as.matrix.coo","matrix.csr",as.matrix.coo.matrix.csr)
+setMethod("as.matrix","matrix.csr",as.matrix.matrix.csr)
+setMethod("as.matrix","matrix.csc",as.matrix.matrix.csc)
+setMethod("as.matrix","matrix.ssr",as.matrix.matrix.ssr)
+setMethod("as.matrix","matrix.ssc",as.matrix.matrix.ssc)
+setMethod("as.matrix","matrix.coo",as.matrix.matrix.coo)
+setMethod("t","matrix.csr",t.matrix.csr)
+setMethod("t","matrix.csc",t.matrix.csc)
+setMethod("t","matrix.coo",function(x) as.matrix.coo(t(as.matrix.csr(x))))
+setMethod("dim","matrix.csr",dim.matrix.csr)
+setMethod("dim","matrix.csc",dim.matrix.csc)
+setMethod("dim","matrix.ssr",dim.matrix.ssr)
+setMethod("dim","matrix.ssc",dim.matrix.ssc)
+setMethod("dim","matrix.coo",dim.matrix.coo)
+setMethod("diff","matrix.csr",diff.matrix.csr)
+setGeneric("diag")
+setMethod("diag","matrix.csr",diag.matrix.csr)
+setMethod("diag<-","matrix.csr",diag.assign.matrix.csr)
+setGeneric("chol")
+setMethod("chol","matrix",chol.default)
+setMethod("chol","matrix.csr",chol.matrix.csr)
+setMethod("chol","matrix.csc",chol.matrix.csc)
+setMethod("backsolve","matrix.csr.chol",backsolve.matrix.csr.chol)
+setMethod("solve","matrix.csr",solve.matrix.csr)
+setMethod("model.matrix","matrix.csc.hb",model.matrix.matrix.csc.hb)
+setMethod("model.matrix","matrix.ssc.hb",model.matrix.matrix.ssc.hb)
+setMethod("model.response",signature(data="matrix.csc.hb",type="ANY"),model.response.matrix.csc.hb)
+setMethod("model.response",signature(data="matrix.ssc.hb",type="ANY"),model.response.matrix.csc.hb)
+setMethod("%*%",signature(x="matrix.csr",y="matrix.csr"),.matmul.matrix.csr)
+setMethod("%*%",signature(x="matrix.csr",y="matrix"),.matmul.matrix.csr)
+setMethod("%*%",signature(x="matrix.csr",y="numeric"),.matmul.matrix.csr)
+setMethod("%*%",signature(x="matrix",y="matrix.csr"),.matmul.matrix.csr)
+setMethod("%*%",signature(x="numeric",y="matrix.csr"),.matmul.matrix.csr)
+setMethod("%x%",signature(X="matrix.csr",Y="matrix.csr"),.kronecker.matrix.csr)
+setMethod("%x%",signature(X="matrix.csr",Y="matrix"),.kronecker.matrix.csr)
+setMethod("%x%",signature(X="matrix.csr",Y="numeric"),.kronecker.matrix.csr)
+setMethod("%x%",signature(X="matrix",Y="matrix.csr"),.kronecker.matrix.csr)
+setMethod("%x%",signature(X="numeric",Y="matrix.csr"),.kronecker.matrix.csr)
+setMethod("image","matrix.csr",image.matrix.csr)
+setMethod("summary","slm",summary.slm)
+setMethod("summary","mslm",summary.mslm)
+setMethod("coef","slm",coef.slm)
+setMethod("fitted","slm",fitted.slm)
+setMethod("residuals","slm",residuals.slm)
+setMethod("print","summary.slm",print.summary.slm)
+#--------------------------------------------------------------------
